#!/usr/bin/ruby
#
# Turing -- Ruby implementation of Captcha
#
# Copyright © 2005 Michal Šafránek <wejn@box.cz>
#
# This file is part of http://turing.rubyforge.org/
#
# See turing.rb in lib/ directory for license terms.
#
require 'cgi'
require 'turing'
require 'turing/challenge'
require 'erb'

# == Generic (Fast)CGI handler to ease integration into current systems
#
# Handles CGI requests using prepared HTML templates and Turing::Challenge
# object.
#
# Example of use:
#   #!/usr/bin/ruby
#   require 'rubygems'
#   gem 'turing'
#   require 'turing'
#   tcgi_config = {
#       :imagepath => "/imgs",
#       :outdir => '/home/wejn/ap/htdocs/imgs',
#       :store => '/home/wejn/ap/data/turing.pstore',
#       :redirect_to => 'http://localhost:8000/secured/',
#   }
#   tcgi_config[:on_success] = proc do
#       out = {}
#       out[:headers] = {
#           "cookie" => CGI::Cookie.new({
#               'name' => 'turing_passed',
#               'value' => 'true',
#               'path' => '/',
#               'expires' => Time.now + 3600*24,
#               }),
#           "dude" => "you_rock!",
#       }
#       out
#   end
#   Turing::CGIHandler.new(tcgi_config).handle
# This CGI script forces user to pass turing challenge. After he passes
# the test, he's given cookie +turing_passed+ with value +true+ and redirected
# to http://localhost:8000/secured/.
#
# <b>Please note</b>: Using this script verbatim is like having no turing
# challenge at all -- any non-braindead attacker will get around it in no
# time.
class Turing::CGIHandler # {{{
	# Configure instance using options hash.
	#
	# *Warning*: Keys of this hash must be symbols.
	#
	# Accepted options:
	# * +templatedir+: directory where templates (challenge.rhtml, error.rhtml, success.rhtml) reside, by default it's set to gem's <tt>shared/templates/</tt> directory.
	# * +on_success+: proc object (or any object responding to +call+) which returns Hash. Recognized keys of the returned hash are +headers+ and +variables+. Former are used as HTTP response headers, latter are used as variables for success template (if no redirect given).
	# * +redirect_to+: redirect to this location on success. This can be also done by returning +Location+ header in +on_success+'s +headers+ hash.
	# * +imagepath+: path under which generated images are accessible.
	#
	# Given hash will be also used to initialize Turing::Challenge object.
	def initialize(options = {}) # {{{
		@options = options
		base = File.join(File.dirname(__FILE__), '..', '..', 'shared')
		
		if @options[:templatedir].nil?
			@options[:templatedir] = File.join(base, 'templates')
		end

		begin
			td = @options[:templatedir]
			raise "not directory" unless FileTest.directory?(td)
		rescue
			error_response("Templatedir is invalid", $!)
		end
		
		begin
			@tc = Turing::Challenge.new(@options)
		rescue
			error_response("Unable to initialize Turing::Challenge class", $!)
			exit 1
		end
	end # }}}
	
	# Handle incoming CGI request.
	#
	# If you don't pass one, it will create it using CGI.new
	def handle(cgi = nil) # {{{
		cgi ||= CGI.new

		if "POST" == cgi.request_method
			if @tc.valid_answer?(cgi["id"], cgi["solution"])
				# process on_success
				os = @options[:on_success]
				if os.nil? || ! os.respond_to?(:call)
					extra = verify_extra({}) # assure correct structure
				else
					extra = verify_extra(os.call)
				end

				# is there redir from config or from on_success?
				if @options[:redirect_to] || extra[:headers]["Location"]
					redirect_to(cgi, extra[:headers]["Location"] || \
						@options[:redirect_to], extra)
				else
					success_response(cgi, extra)
				end
			else
				show_challenge(cgi, true)
			end
		else
			show_challenge(cgi)
		end
	end # }}}

	private

	# generate error exception
	def error_response(what, exc, cgi = nil) # {{{
		cgi ||= CGI.new
		$stderr.puts "What: #{$!.to_s}"
		cgi.out do
			begin
				Template.new(template_file("error.rhtml"), {
					:what => what,
					:explanation => exc.to_s,
					:backtrace => exc.backtrace.join("\n"),
				}).render
			rescue
				err = [what, '', exc.to_s, '', exc.backtrace.join("\n")]
				"<pre>" + CGI.escapeHTML(err.join("\n")) + "</pre>"
			end
		end
	end # }}}

	# generate challenge and show it
	def show_challenge(cgi, bad_try = false) # {{{
		c = nil
		begin
			c = @tc.generate_challenge
		rescue
			error_response("Unable to generate challenge", $!, cgi)
		end

		cgi.out do
			begin
				Template.new(template_file("challenge.rhtml"), {
					:already_failed => bad_try,
					:file => File.join(@options[:imagepath] || '', c.file),
					:id => c.id,
				}).render
			rescue => exc
				error_response("Unable to render template", exc, cgi)
			end
		end
	end # }}}

	# show success response
	def success_response(cgi, extra = {}) # {{{
		cgi.out(extra[:headers]) do
			begin
				Template.new(template_file("success.rhtml"),
					extra[:variables]).render
			rescue => exc
				error_response("Unable to render template", exc, cgi)
			end
		end
	end # }}}

	# redirect to given location
	def redirect_to(cgi, where, extra = {}) # {{{
		params = {
			"Location" => where,
		}
		params.merge!(extra[:headers])
		cgi.out(params) do
			<<-EOF
<!DOCTYPE HTML PUBLIC "-//IETF//DTD HTML 2.0//EN">
<HTML><HEAD>
<META HTTP-EQUIV='refresh' CONTENT='1;url=#{CGI.escapeHTML(where.to_s)}'>
<TITLE>302 Found</TITLE>
</HEAD><BODY>
<H1>Found</H1>
The document has moved <A HREF="#{CGI.escapeHTML(where.to_s)}">here</A>.<P>
</BODY></HTML>
			EOF
		end

	end # }}}

	# verify/correct "extra" object returned from on_success
	def verify_extra(extra) # {{{
		out = {
			:headers => {},
			:variables => {},
		}
		extra = {} if extra.nil? || ! extra.kind_of?(Hash)
		out[:headers] = extra[:headers] if extra[:headers].kind_of?(Hash)
		out[:variables] = extra[:variables] if extra[:variables].kind_of?(Hash)
		out
	end # }}}
	
	# template name to abs path to file
	def template_file(name) # {{{
		name += ".rhtml" unless name =~ /\.rhtml$/
		File.join(@options[:templatedir], name)
	end # }}}

	public

	# == Simple template engine
	# Essentially just convenient wrapper around ERB which is used internally by Turing::CGIHandler.
	class Template # {{{
		# Specify template file (+what+) and +variables+ that will be pushed as instance variables to the Template.
		def initialize(what, variables = nil) # {{{
			(variables || {}).each do |k,v|
				instance_variable_set("@" + k.to_s, v)
			end
			@__what__ = what
		end # }}}

		# render given template and return result as string.
		def render # {{{
			erb = ERB.new(File.open(@__what__).read, nil, '%-')
			erb.result(binding)
		rescue
			raise "Failure rendering template #{@what}: #{$!}"
		end # }}}

		# shortcut for <tt>CGI.escapeHTML</tt>
		#
		# <i>can you say "Rails" ? :)</i>
		def h(var) # {{{
			CGI.escapeHTML(var)
		end # }}}
	end # }}}
end # }}}

# execute if called directly
if __FILE__ == $0 # {{{
	Turing::CGIHandler.new({
		:imagepath => '.',
		:outdir => '.',
		:store => 'store',
	}).handle
	exit 0 # 'turing' requires 'turing/challenge' which might make loop
end # }}}

# vim: set ts=4 sw=4 :
